/*
 * Copyright (C) 2002-2003  Pekka Enberg <Pekka.Enberg@cs.helsinki.fi>
 *
 * Distributed under the terms of the GNU General Public License
 * version 2 or later.
 */
#include <assert.h>
#include <curses.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include "error.h"
#include "colors.h"

/*
 * Off-screen edit buffer
 */
struct edit_buffer {
  unsigned long start_x;
  unsigned long start_y;
  unsigned long height;
  unsigned long width;
  unsigned int * buffer;
};

/*
 * Visible screen
 */
struct screen {
  unsigned long cursor_x;
  unsigned long cursor_y;
  unsigned long height;
  unsigned long width;
};

struct editor_context {
  int fg_color;
  int bg_color;
};

void
set_attribute (int attribute)
{
  int fg_color = (attribute & 0x0F);
  int bg_color = (attribute & 0xF0) >> 4;

  /* We need to turn on the BOLD attribute to get colors > 7.  */
  if (fg_color > 7)
    {
      attron (A_BOLD);
      fg_color -= 7;
    }
  else
    {
      attroff (A_BOLD);
    }
  attron (COLOR_PAIR (attr_to_color_pair(fg_color, bg_color)));
}

#define COLOR_ATTR(fg, bg) (bg * 0x10 + fg)

/*
 * If you use any of the colors, please define a constant here instead
 * of hardcoding the value into the code.
 */
#define RED_ON_BLACK 1

void
print_status (struct edit_buffer * buf,
	      struct screen * scr,
	      struct editor_context * ctx)
{
  /*
   * Last line on screen is the status bar
   */
  move (scr->height, 0);
  clrtoeol ();

  attron (COLOR_PAIR(RED_ON_BLACK));
  mvprintw (scr->height, 1, "(%2i, %2i)",
	    scr->cursor_x + buf->start_x + 1,
	    scr->cursor_y + buf->start_y + 1);
  attroff (COLOR_PAIR(RED_ON_BLACK));

  move (scr->height, 13);
  set_attribute (COLOR_ATTR (ctx->fg_color, ctx->bg_color));
  printw ("Color");
}

void
edit_buffer_put (struct edit_buffer * buf, unsigned long x, unsigned long y, int value)
{
  assert (x < buf->width ); assert (x >= 0);
  assert (y < buf->height); assert (y >= 0);

  buf->buffer[y * buf->width + x] = value;
}

int
edit_buffer_get (struct edit_buffer * buf, unsigned long x, unsigned long y)
{
  assert (x < buf->width ); assert(x >= 0);
  assert (y < buf->height); assert(y >= 0);

  return buf->buffer[y * buf->width + x];
}

void
edit_buffer_clear (struct edit_buffer * buf)
{
  int x, y;

  for (y = 0; y < buf->height; y++)
    for (x = 0; x < buf->width; x++)
      edit_buffer_put (buf, x, y, 0x0720);
}

void
edit_buffer_draw_to_screen (struct edit_buffer * buf, struct screen * scr)
{
  assert (buf->start_x + scr->width  <= buf->width );
  assert (buf->start_y + scr->height <= buf->height);
  assert (buf->start_x >= 0);
  assert (buf->start_y >= 0);

  int x, y;

  for (y = 0; y < scr->height; y++)
    for (x = 0; x < scr->width; x++)
      {
	int attribute = (edit_buffer_get (buf,
					  buf->start_x + x,
					  buf->start_y + y) & 0xFF00) >> 8;
	int character = edit_buffer_get (buf,
					 buf->start_x + x,
					 buf->start_y + y) & 0xFF;
	set_attribute (attribute);
	mvprintw (y, x, "%c", character);
      }
}

void
edit_buffer_init (struct edit_buffer * buf, unsigned long width, unsigned long height)
{
  buf->buffer = malloc(height * width * sizeof(int));
  if (buf->buffer == NULL)
    {
      error("Could not allocate memory for edit buffer.");
    }
  buf->height = height;
  buf->width = width;
}

void
edit_buffer_release (struct edit_buffer * buf)
{
  if (buf->buffer != NULL)
    free(buf->buffer);

  buf->buffer = NULL;
  buf->width = 0;
  buf->height = 0;
}

/* Curses type KEY_xxx check if META key was used for character.  */
#define KEY_META(ch) (0x1000 | ch)

/* ALT and ESC are both the same META key for terminals.  */
#define META_KEY_CODE 0x1B

int
get_char ()
{
  int ch = getch ();
  if (ch == META_KEY_CODE)
    {
      /* After META comes the actual key we're interested in.  */
      ch = KEY_META(getch());
    }
  return ch;
}

int
get_printable_char (int ch)
{
  /* TODO: Function keys should go trough lookup table.  */
  switch (ch)
    {
    case KEY_F(1):
      return 176;
    case KEY_F(2):
      return 177;
    case KEY_F(3):
      return 178;
    case KEY_F(4):
      return 219;
    case KEY_F(5):
      return 220;
    case KEY_F(6):
      return 223;
    case KEY_F(7):
      return 221;
    case KEY_F(8):
      return 222;
    case KEY_F(10):
      return 254;
    }
  return ch;
}

void
screen_init (struct screen * scr)
{
  getmaxyx (stdscr, scr->height, scr->width);
  if (scr->height == -1 || scr->width == -1)
    error("Could not get screen dimensions from curses.");
  
  /* Leave a line for the status bar.  */
  scr->height--;
}

/*
 * Commands issued in editor
 */

void
cmd_resize ()
{
  /* Currently we don't support resizing the terminal while
     running.  */
  error ("Terminal was resized.");
}

void
cmd_move_up (struct edit_buffer * buf, struct screen * scr)
{
  if (scr->cursor_y > 0)
    scr->cursor_y--;
  else if (buf->start_y > 0)
    buf->start_y--;
}

void
cmd_move_down (struct edit_buffer * buf, struct screen * scr)
{
  if (scr->cursor_y < (scr->height - 1))
    scr->cursor_y++;
  else if (buf->start_y + scr->height < buf->height)
    buf->start_y++;
}

void
cmd_move_left (struct edit_buffer * buf, struct screen * scr)
{
  if (scr->cursor_x > 0)
    scr->cursor_x--;
  else if (buf->start_x > 0)
    buf->start_x--;
}

void
cmd_move_right (struct edit_buffer * buf, struct screen * scr)
{
  if (scr->cursor_x < (scr->width - 1))
    scr->cursor_x++;
  else if (buf->start_x < (buf->width - scr->width))
    buf->start_x++;
}

void
cmd_move_to_end (struct edit_buffer * buf, struct screen * scr)
{
  scr->cursor_x = (scr->width - 1);
  buf->start_x = buf->width - scr->width;
}

void
cmd_move_to_start (struct edit_buffer * buf, struct screen * scr)
{
  scr->cursor_x = 0;
  buf->start_x = 0;
}

void
cmd_print_char (struct edit_buffer * buf,
		struct screen * scr,
		struct editor_context * ctx,
		int ch)
{
  edit_buffer_put (buf,
		   buf->start_x + scr->cursor_x,
		   buf->start_y + scr->cursor_y,
		   (COLOR_ATTR (ctx->fg_color, ctx->bg_color) << 8)
		   | get_printable_char (ch));
}

void
cmd_move_page_up (struct edit_buffer * buf, struct screen * scr)
{
  if (buf->start_y > scr->height)
    {
      buf->start_y -= scr->height;
    }
  else
    {
      buf->start_y = 0;
      scr->cursor_y = 0;
    }
}

void
cmd_move_page_down (struct edit_buffer * buf, struct screen * scr)
{
  if (buf->start_y + (scr->height * 2) < buf->height)
    {
      buf->start_y += scr->height;
    }
  else
    {
      buf->start_y = buf->height - scr->height;
      scr->cursor_y = (scr->height - 1);
    }
}

#define MAX_BG_COLOR 7
#define MAX_FG_COLOR 14

#define DECREASE_AND_WRAP(c, max) (c >   0 ? c - 1 : max)
#define INCREASE_AND_WRAP(c, max) (c < max ? c + 1 :   0)

void
cmd_next_fg_color (struct editor_context * ctx)
{
  ctx->fg_color = INCREASE_AND_WRAP (ctx->fg_color, MAX_FG_COLOR);
}

void
cmd_prev_fg_color (struct editor_context * ctx)
{
  ctx->fg_color = DECREASE_AND_WRAP (ctx->fg_color, MAX_FG_COLOR);
}

void
cmd_next_bg_color (struct editor_context * ctx)
{
  ctx->bg_color = INCREASE_AND_WRAP (ctx->bg_color, MAX_BG_COLOR);
}

void
cmd_prev_bg_color (struct editor_context * ctx)
{
  ctx->bg_color = DECREASE_AND_WRAP (ctx->bg_color, MAX_BG_COLOR);
}

/*
 * Main editor loop
 */

void
edit_loop ()
{
  struct screen _scr = {
    .cursor_x = 0,
    .cursor_y = 0
  };

  struct editor_context _ctx = {
    .fg_color = 0x07,
    .bg_color = 0x00
  };

  struct edit_buffer _buf = {
    .start_x = 0,
    .start_y = 0
  };

  struct screen * scr = &_scr;
  struct editor_context * ctx = &_ctx;
  struct edit_buffer * buf = &_buf;

  screen_init (scr);
  edit_buffer_init (buf, 160, 100);
  edit_buffer_clear (buf);

  bool quit = false;
  while (!quit)
    {
      edit_buffer_draw_to_screen (buf, scr);
      print_status (buf, scr, ctx);
      move (scr->cursor_y, scr->cursor_x);

      int ch = get_char ();
      if (ch == ERR)
	{
	  error ("getch() returned ERR");
	}
      switch (ch)
	{
	case KEY_META('x'):
	case KEY_META('X'): quit = true; break;

	case KEY_RESIZE: cmd_resize (); break;
	case KEY_UP:     cmd_move_up (buf, scr); break;
	case KEY_DOWN:   cmd_move_down (buf, scr); break;
	case KEY_LEFT:   cmd_move_left (buf, scr); break;
	case KEY_RIGHT:  cmd_move_right (buf, scr); break;
	case KEY_END:    cmd_move_to_end (buf, scr); break;
	case KEY_HOME:   cmd_move_to_start (buf, scr); break;
	case KEY_PPAGE:  cmd_move_page_up (buf, scr); break;
	case KEY_NPAGE:  cmd_move_page_down (buf, scr); break;

	case KEY_META(KEY_UP):    cmd_next_bg_color (ctx); break;
	case KEY_META(KEY_DOWN):  cmd_prev_bg_color (ctx); break;
	case KEY_META(KEY_RIGHT): cmd_next_fg_color (ctx); break;
	case KEY_META(KEY_LEFT):  cmd_prev_fg_color (ctx); break;

	case KEY_BACKSPACE:
	  cmd_move_left (buf, scr);
	  cmd_print_char (buf, scr, ctx, ' ');
	  break;

	default:
	  cmd_print_char (buf, scr, ctx, ch);
	  cmd_move_right (buf, scr);
	}
    }
  edit_buffer_release (buf);
}

void
init_curses ()
{
  savetty ();
  initscr ();
  nonl ();
  cbreak ();
  noecho ();
  keypad (stdscr, TRUE);

  if (has_colors() == FALSE)
    {
      error("Your terminal doesn't support colors.");
    }
  start_color ();
}

void
release_curses ()
{
  resetty ();
  endwin ();
}

int
main ()
{
  init_curses ();
  init_color_pairs ();
  edit_loop();
  release_curses ();
  printf ("Thank you for using NewDraw.\n");

  return EXIT_SUCCESS;
}

/*
 * Local Variables:
 * c-file-style: "gnu"
 * End:
 */
